####################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

from typing import Set, Tuple, List, Dict
from abc import *

from laygo2.object.template import UserDefinedTemplate
import laygo2.object.database

import numpy as np


class TileTemplate(UserDefinedTemplate, metaclass = ABCMeta):
    """ 
    The class for basic parametric cell. 
    The sub-elements will be placed by placement_pattern.

    generation flow:
        1. insert parameters by __init__()
        2. use self.generate_func, self.generete_pin, self.bbox for creating UserDefindTemplate
    """

    glib         = None  
    """The main grid libarary for routing."""
    
    tlib         = None  
    """The main template library for calling nativeTemplates."""
    
    placement_map: Dict[str, str]  = None  
    """The mapping of the name of sub-elements and the name of native template.
       the name of sub_elements = {"core", "bndr", "bndl", "gbndr", "gbndl"}.
    """

    placement_pattern: List[str] = None
    """The sequence of sub-elements placement."""
      
    transform_pattern: List[str] = None            
    """The sequencye of sub-elements transform: {"R0", "MX", "MY", R180"}."""
    
    routing_map: Dict[str, int]  = None  
    """The mapping of Mosfet terminal and r2 vertical track and other options for tracks.
       Mosfet terminal = {"G", "D", "S"},
       
    """
    
    routing_gname: str         = None  
    """The name of main routing grid."""

    nf_core: int               = None  
    """The core template number of finger."""

    libname:str                = None 
    """The name of Native template library."""

    def __init__(self, tlib, glib, routing_gname: str, routing_map: dict, placement_map: dict, placement_pattern: list, transform_pattern: list, name: str):
        
        self.tlib              = tlib
        self.glib              = glib
        self.routing_gname     = routing_gname     # mosfet grid name
        self.routing_map       = routing_map       # gate, drain, source trakcs
        self.placement_map     = placement_map     # native template name dictionary
        self.placement_pattern = placement_pattern # placement pattern
        self.transform_pattern = transform_pattern # transform pattern
        
        self.libname = tlib.libname
        self.nf_core   = 2
        
        super().__init__(name = name, bbox_func = self.bbox_func, pins_func = self.pins_func, generate_func = self.generate_func)

    def _route_pattern_via_track(self, grid, mn_list, n_track, name_via ) -> list:
        """ Route with multi instance via

        Parameters
        ----------
        mn_list: np.array 
            sorted and unique array
        n_track: int
            track number of n
        name_via: str
            the name of via
        """
        len_via = mn_list.shape[0]
        m_left  = mn_list[0][0] # left-most m
        tvia    = grid.viamap[m_left, n_track]
            
        if len_via == 1:
            via     = tvia.generate(name = name_via, shape = ( len_via, 1), pitch = [1, 1] )
            track   = None
        else:
            m_left2   = mn_list[1][0]
            pitch_m   = m_left2 - m_left
            pitch_x   = grid.x(pitch_m)

            via       = tvia.generate(name = name_via, shape = ( len_via, 1), pitch = [pitch_x, 0])
            track     = grid.route( mn = ( [mn_list[0][0], n_track ] , [mn_list[-1][0], n_track ]   ) )
            
        via.xy = grid.xy( [m_left, n_track ])

        return via, track

    def _update_params(self, params_in) -> dict:
        """ Update sub-elements parameters
        """

        params = {}

        if "switch" in params_in:
            nfs               = dict(core = "nf", dmyl = "nfdmyl", dmyr = "nfdmyr")
            flags             = params_in["switch"]
            sizes             = params_in["size"]
            placement_pattern = self.placement_pattern

            for i in len(flags):
                tname = placement_pattern[i]
                size  = sizes[i]
                flag  = flags[i] 
                if flag == True:
                    params_in[tname] = True
                    if tname in nfs:
                        params_in[ nfs[tname] ] = size   
                else:
                    params_in[tname] = False

        params["nf"]        = params_in.get("nf"     , self.nf_core )
        params["nfdmyl"]    = params_in.get("nfdmyl" , False )
        params["nfdmyr"]    = params_in.get("nfdmyr" , False )
        
        params["dmyl"] = True
        params["dmyr"] = True

        if params["nfdmyl"] == False:
            params["dmyl"]      = False

        if params["nfdmyr"] == False:
            params["dmyr"]      = False

        params["core"]      = params_in["nf"]

        params["bndl" ]     = params_in.get("bndl"   , True )
        params["bndr" ]     = params_in.get("bndr"   , True )
        
        params["gbndl" ]     = params_in.get("gbndl" , False)
        params["gbndr" ]     = params_in.get("gbndr" , False)
        
        params = self._update_params_sub(params_in, params)
        return params
    
    @abstractmethod
    def _update_params_sub(self, params_in, params):
        """ update custom flags """
        pass

    @abstractmethod
    def _internal_route(self, params):
        """ internal routing method """
        pass

    @abstractmethod
    def pins_func(self, params):
        """ pin generation method """
        pass

    def _generate_subinstance(self, params) -> dict:
        """The method of generating native_elements.
        """

        tlib            = self.tlib
        placement_map   = self.placement_map
        params    = self._update_params(params)
        nf_core   = self.nf_core
        tfs       = self.transform_pattern
        
        iparams   = dict()  # mapping sub-elements and generated instance.
        nf        = int( params["nf"] / nf_core) 
        
        if params["nfdmyl"] == False:
            nfdmyl = 1
        else:
            nfdmyl = params["nfdmyl"]

        if params["nfdmyr"] == False:
            nfdmyr = 1
        else:
            nfdmyr = params["nfdmyr"]

        tf_set = ("R0", "MY")

        if nf_core == 1: # not used, source and drain swap continously
            icore_sub = [0]* nf
            nelements_core = {}
            pins_core      = {}
            for i in range( nf):
                tf_core   = tf_set[ i%2]
                icore_sub = tlib[ placement_map["core"] ].generate( name=f'IM{i}'   , shape = [1, 1], transform = tf_core )
                nelements_core[f"IM{i}"] = icore_sub
                pins_core[f"D{i}"]   = icore_sub.pins["D"]
                pins_core[f"S{i}"]   = icore_sub.pins["S"]
                pins_core[f"G{i}"]   = icore_sub.pins["G"]
            icore  = laygo2.object.physical.VirtualInstance( xy = [0,0], libname = "dummy", cellname="dummy",
                                                             pins = pins_core,
                                                             name = "IM0", native_elements = nelements_core, transform = "R0"
                                                            )
        else:
            icore  = tlib[ placement_map["core"] ].generate( name='ICORE0', shape = [nf, 1] , transform = tfs["core"] )
        
        ibndl  = tlib[ placement_map["bndl"] ].generate( name='IBNDL0', shape = [1, 1] , transform = tfs["bndl"] )
        ibndr  = tlib[ placement_map["bndr"] ].generate( name='IBNDR0', shape = [1, 1] , transform = tfs["bndr"] )
        
        idmyl  = tlib[ placement_map["dmyl"] ].generate( name='IDMYL0',  shape = [max( int(0.5 * nfdmyl ),1)  , 1], transform = tfs["dmyl"] )
        idmyr  = tlib[ placement_map["dmyr"] ].generate( name='IDMYR0',  shape = [max( int(0.5 * nfdmyr) ,1)  , 1], transform = tfs["dmyr"] )

        igbndl = tlib[ placement_map["gbndl"] ].generate( name='IGDMYL0', shape = [1, 1], transform = tfs["gbndl"] )
        igbndr = tlib[ placement_map["gbndr"] ].generate( name='IGDMYR0', shape = [1, 1], transform = tfs["gbndr"] )

        iparams["gbndl"] = igbndl
        iparams["gbndr"] = igbndr
        
        iparams["dmyl"]  = idmyl
        iparams["dmyr"]  = idmyr

        iparams["bndl"]  = ibndl
        iparams["bndr"]  = ibndr
        
        iparams["core"]  = icore

        return iparams

    def _device_place(self, params) -> dict:
        """ The method of placing native elements.
            It will modify instances.xy
        """

        params        = self._update_params( params )
        iparams       = self._generate_subinstance( params )
        
        placement_pattern = self.placement_pattern
        tfs               = self.transform_pattern

        # placement
        cursor = [0, 0]
        for i, t_inst in enumerate( placement_pattern ):
            if params[t_inst] != False:
                inst     = iparams[t_inst]
            
                if tfs[t_inst] == "MY":
                    xy_bl    = inst.bbox[0]  # this is on the left side
                    xy_tr    = inst.bbox[1]
                    inst.xy  = np.asarray( cursor ) + [ -xy_bl[0], 0 ]
                else:
                    inst.xy  = cursor
            
                xy_bl    = inst.bbox[0]
                xy_tr    = inst.bbox[1]
                cursor   = [ xy_tr[0], xy_bl[1] ] 
            else:
                iparams.pop(t_inst) # remove from dictionary
        
        return iparams

   
    def bbox_func(self, params) -> np.ndarray:
        params  = self._update_params(params)
        
        iparams = self._device_place( params )

        xy_all = np.asarray([0,0])
        for t_inst, inst in iparams.items():
            xy = inst.bbox
            xy_bl     = xy[0]
            xy_tr     = xy[1]

            xy_all[0] = max( xy_all[0], xy_tr[0])
            xy_all[1] = max( xy_all[1], xy_tr[1])
        
        return xy_all
    
    def generate_func(self, name = None, shape = None, pitch = None, transform = 'R0', params = None):
        """ generate Virtualinstances and with routing
        """

        nelements = dict()
        params    = self._update_params( params )
        placement_map   = self.placement_map
        libname   = self.libname
        iparams   = self._device_place( params )
        nelements.update( iparams )            
        
        routes    = self._internal_route( params )
        nelements.update( routes )            

        pins      = self.pins_func(params)
        nelements.update(pins)

        bbox      = self.bbox_func(params)

        # Generate and return the final instance
        inst = laygo2.object.VirtualInstance( name = name, xy = np.array([0, 0]) , libname = libname, cellname = f'myvcell_{name}' ,
                                              native_elements = nelements, shape = shape, pitch = pitch,
                                              transform       = transform, unit_size = bbox, pins = pins)
        
        return inst

class TileMOSTemplate(TileTemplate):

    def _update_params_sub(self, params_in, params):
        # about routing
        params["sdswap"]    = params_in.get("sdswap"   , False)  
        params["trackswap"] = params_in.get("trackswap", False)  
        params["tie"      ] = params_in.get("tie"      , False)  
        
        # new function
        params["rtrackswap"] = params_in.get("rtrackswap", False)  
        params["sdswap"]     = params_in.get("sdswap"    , False)  
        params["rail"     ]  = params_in.get("rail", True )  
        
        return params

    def pins_func(self, params):
        """The method of creating pin."""
        params  = self._update_params(params)
        pins    = dict()
        
        # generate a virtual routing structure for reference
        route_obj = self._internal_route( params )

        if 'RG0' in route_obj:  # gate
            g_obj = route_obj['RG0']
            pins['G'] = laygo2.object.Pin(xy=g_obj.xy, layer=g_obj.layer, netname='G')
        if 'RD0' in route_obj:  # drain
            d_obj = route_obj['RD0']
            pins['D'] = laygo2.object.Pin(xy=d_obj.xy, layer=d_obj.layer, netname='D')
        if 'RS0' in route_obj:  # source
            s_obj = route_obj['RS0']
            pins['S'] = laygo2.object.Pin(xy=s_obj.xy, layer=s_obj.layer, netname='S')
        if 'RRAIL0' in route_obj:  # rail
            r_obj = route_obj['RRAIL0']
            pins['RAIL'] = laygo2.object.Pin(xy=r_obj.xy, layer=r_obj.layer, netname='RAIL')
        return pins

    def _internal_route(self, params) -> dict:
        """The method of routing mosfet"""

        params            = self._update_params(params)
        glib              = self.glib
        routing_gname     = self.routing_gname
        routing_map       = self.routing_map
        nf_core           = self.nf_core
        iparams           = self._device_place(params)
        placement_pattern = self.placement_pattern

        r12 = glib[routing_gname]
        n_g = routing_map["G"]
        n_d = routing_map["D"]
        n_s = routing_map["S"]
        n_r = routing_map.get( "RAIL", 0 )

        nelements = {}
        nelements.update(iparams)
        
        icore   = iparams["core"]
        nf      = params["nf"]

        G, S, D = "G", "S", "D"
        
        if params["trackswap"]: #backward compatable
            n_d, n_s = n_s, n_d
            S, D     = D, S
        
        if params["rtrackswap"]: 
            n_d, n_s = n_s, n_d
        
        if params["sdswap"]: 
            S, D     = D, S

        # Extrack mn from core mosfet
        mn_list_g = [] 
        mn_list_d = [] 
        mn_list_s = []
        i_iter = int( nf / nf_core) 

        for i in range( i_iter ):
            if i_iter == 1:
                icore_sub =  icore
            else:
                icore_sub =  icore[i][0] # it is 2-dimensional array
                
            for pin_name in icore_sub.pins.keys():

                if G in pin_name:
                    mn_list_g.append( r12.mn.center( icore_sub.pins[pin_name] ) )
                elif D in pin_name:
                    mn_list_d.append( r12.mn.center( icore_sub.pins[pin_name] ) )
                elif S in pin_name:
                    mn_list_s.append( r12.mn.center( icore_sub.pins[pin_name] ) )
        

        mn_list_g = np.asarray( mn_list_g)
        mn_list_g = mn_list_g[  mn_list_g[:,0].argsort()] # sort list by bottom_left
        mn_list_g = np.unique(  mn_list_g, axis = 0 )

        mn_list_d = np.asarray( mn_list_d)
        mn_list_d = mn_list_d[  mn_list_d[:,0].argsort()] 
        mn_list_d = np.unique(  mn_list_d, axis = 0 )
        
        mn_list_s = np.asarray( mn_list_s)
        mn_list_s = mn_list_s[  mn_list_s[:,0].argsort()] 
        mn_list_s = np.unique(  mn_list_s, axis = 0 )

        # Create via and track
        r_g = dict( via = 0 , track = 0) 
        r_d = dict( via = 0 , track = 0) 
        r_s = dict( via = 0 , track = 0) 

        for mn_list, n_track, terminal, r_dict in zip( (mn_list_g, mn_list_d, mn_list_s), (n_g, n_d, n_s), (G, D, S), (r_g, r_d, r_s) ):
            
            via, track = self._route_pattern_via_track( r12, mn_list, n_track, terminal)
            r_dict["via"]   = via
            r_dict["track"] = track 

        # Modify track
        if r_g["track"] == None: 
            mn_sub       = mn_list_g[0]
            r_g["track"] = r12.route( mn = [ mn_sub + [0,0], mn_sub + [0,0] ], via_tag =[None, None] )  # new track
            
            x_extl        = routing_map["G_extension0_x"][0]
            x_extr        = routing_map["G_extension0_x"][1]
            if x_extl != None:
                xy_sub     = r12.xy(mn_sub)
                r_g["track"].xy = [ xy_sub + [ x_extl , 0], xy_sub + [x_extr, 0] ] # new track location

        if r_d["track"] == None: 
            mn_sub    = mn_list_d[0]
            mn_sub[1] = n_d
            
            m_ext = routing_map["D_extension0_m"]
            m_extl = 0
            m_extr = 0
            if m_ext[0] != None:
                m_extl = m_ext[0]
                m_extr = m_ext[1]

            r_d["track"]   = r12.route( mn = [ mn_sub + [m_extl,0], mn_sub + [m_extr,0] ], via_tag =[None, None] )

        if r_s["track"] == None: 
            mn_sub    = mn_list_s[0]
            mn_sub[1] = n_s
            
            m_ext = routing_map["S_extension0_m"]
            m_extl = 0
            m_extr = 0
            if m_ext[0] != None:
                m_extl = m_ext[0]
                m_extr = m_ext[1]
            r_s["track"]   = r12.route( mn = [ mn_sub + [m_extl,0], mn_sub + [m_extr,0] ], via_tag =[None, None] )

        nelements["VIA_G"] = r_g["via"]
        nelements["VIA_D"] = r_d["via"]
        nelements["VIA_S"] = r_s["via"]

        nelements["RG0"] = r_g["track"]
        nelements["RD0"] = r_d["track"]
        nelements["RS0"] = r_s["track"]

        # Rail routing
        if params["rail"]:

            tinst_l = 0 # left most sub-element name
            tinst_r = 0 # right most sub-element name

            for i, tinst in enumerate(placement_pattern):
                name = placement_pattern[i]
                if name in iparams:
                    if name != "gbndl":
                        tinst_l = name
                        break   
            
            for i, tinst in enumerate(placement_pattern):
                name = placement_pattern[-1 - i]
                if name in iparams:
                    if name != "gbndr":
                        tinst_r = name
                        break   
                    
            inst_l = iparams[tinst_l]
            inst_r = iparams[tinst_r]

            xy_bl = inst_l.bbox[0]
            xy_tr = inst_r.bbox[1]
            xy_br = [xy_tr[0], xy_bl[1] ]

            mn_bl = r12.xy >= xy_bl 
            mn_br = r12.xy <= xy_br 

            r_track = r12.route( mn = [mn_bl, mn_br], via_tag = [None, None] )
            nelements["RRAIL0"] = r_track

        # TIE
        if params["tie"] != False:

            if params["tie"] == "S" or params["tie"] == True:
                del nelements["RS0"]
                del nelements["VIA_S"]
                
                via, track = self._route_pattern_via_track(r12, mn_list_s, n_r, "VIA_TIE")
                r_t        = r12.route_via_track( mn = mn_list_s, via_tag = [ False, False], track = [None, n_r])
            
            elif params["tie"] == "D":
                del nelements["RD0"]
                del nelements["VIA_D"]
                
                via, track = self._route_pattern_via_track(r12, mn_list_d, n_r, "VIA_TIE")
                r_t        = r12.route_via_track( mn = mn_list_d, via_tag = [ False, False], track = [None, n_r])

            else:
                raise Exception(" Value error for tie")
            
            nelements["VIA_TIE"] = via
            
            if r_t[-1] == None:
                del r_t[1]
            for i, r in enumerate(r_t): # source or drain to rail
                nelements[f"RECT_TIE{i}"] = r

        # dmy routing

        names_dmy = ["dmyl", "dmyr"]
        for i, tdmy_nf in enumerate( ( "nfdmyl", "nfdmyr" )):
            if params[tdmy_nf] == 0:
                continue
            name_dmy  = names_dmy[i]
            icore     = iparams[ name_dmy ]
            mn_list_g = [] 
            mn_list_d = [] 
            mn_list_s = []
            i_iter = int( params[tdmy_nf] / nf_core) 

            for i in range( i_iter ):
                if i_iter == 1:
                    icore_sub =  icore
                else:
                    icore_sub =  icore[i][0] # it is 2-dimensional array
                
                for pin_name in icore_sub.pins.keys():

                    if G in pin_name:
                        mn_list_g.append( r12.mn.center( icore_sub.pins[pin_name] ) )
                    elif D in pin_name:
                        mn_list_d.append( r12.mn.center( icore_sub.pins[pin_name] ) )
                    elif S in pin_name:
                        mn_list_s.append( r12.mn.center( icore_sub.pins[pin_name] ) )

                mn_list_a = mn_list_g + mn_list_s + mn_list_d
                mn_list_a = np.asarray( mn_list_a)
                mn_list_a = mn_list_a[  mn_list_a[:,0].argsort()] # sort list by bottom_left
                mn_list_a = np.unique(  mn_list_a, axis = 0 )

                r_t        = r12.route_via_track( mn = mn_list_a, via_tag = [ False, False], track = [None, n_r])
                
                mn_list_a[:,1] = n_r
                mn_list_a = np.unique(  mn_list_a, axis = 0 )
                via, track = self._route_pattern_via_track(r12, mn_list_a, n_r, f"VIA_DUMMY_{icore.name}")
                
                nelements[f"VIA_TIE_{name_dmy}"] = via
                
                if r_t[-1] == None:
                    del r_t[1]
                for i, r in enumerate(r_t): # source or drain to rail
                    nelements[f"RECT_DMY_TIE_{name_dmy}_{i}"] = r

        return nelements
  

class TileTapTemplate(TileTemplate):
    """The class of TileTap."""

    def _update_params_sub(self, params_in, params):
        # about routing
        params["sdswap"]    = params_in.get("sdswap"   , False)  
        params["trackswap"] = params_in.get("trackswap", False)  
        params["tie"      ] = params_in.get("tie"      , False)  
        
        # new function
        params["rtrackswap"] = params_in.get("rtrackswap", False)  
        params["sdswap"]     = params_in.get("sdswap"    , False)  
        params["rail"     ]  = params_in.get("rail", True )  
        
        return params

    def pins_func(self, params):
        """The method of creating pin."""
        params  = self._update_params(params)
        pins    = dict()
        
        # generate a virtual routing structure for reference
        route_obj = self._internal_route( params )

        if 'RG0' in route_obj:  # gate
            g_obj = route_obj['RG0']
            pins['G'] = laygo2.object.Pin(xy=g_obj.xy, layer=g_obj.layer, netname='G')
        if 'RD0' in route_obj:  # drain
            d_obj = route_obj['RD0']
            pins['D'] = laygo2.object.Pin(xy=d_obj.xy, layer=d_obj.layer, netname='D')
        if 'RS0' in route_obj:  # source
            s_obj = route_obj['RS0']
            pins['S'] = laygo2.object.Pin(xy=s_obj.xy, layer=s_obj.layer, netname='S')
        if 'RRAIL0' in route_obj:  # rail
            r_obj = route_obj['RRAIL0']
            pins['RAIL'] = laygo2.object.Pin(xy=r_obj.xy, layer=r_obj.layer, netname='RAIL')
        return pins


    def _internal_route(self, params):
        params    = self._update_params(params)
        glib      = self.glib
        routing_gname = self.routing_gname
        routing_map = self.routing_map
        nf_core   = self.nf_core
        iparams   = self._device_place(params)
        placement_pattern = self.placement_pattern

        r12 = glib[routing_gname]
        n_g = routing_map["G"]
        n_d = routing_map["D"]
        n_s = routing_map["S"]
        n_r = routing_map.get( "RAIL", 0 )

        nelements = {}
        nelements.update(iparams)
        
        icore   = iparams["core"]
        nf      = params["nf"]

        G, S, D = "G", "S", "D"
        
        if params["trackswap"]: #backward compatable
            n_d, n_s = n_s, n_d
            S, D     = D, S
        
        if params["rtrackswap"]: 
            n_d, n_s = n_s, n_d
        
        if params["sdswap"]: 
            S, D     = D, S

        mn_list_s = [] # TAP0
        mn_list_d = [] # TAP1

        i_iter = int( nf / nf_core) 
        for i in range( i_iter ):
            if i_iter == 1:
                icore_sub =  icore
            else:
                icore_sub =  icore[i][0] # it is 2-dimensional array
                
            for pin_name in icore_sub.pins.keys():
                if "TAP0" in pin_name:
                    mn_list_s.append( r12.mn.center( icore_sub.pins[pin_name] ) )
                elif "TAP1" in pin_name:
                    mn_list_d.append( r12.mn.center( icore_sub.pins[pin_name] ) )
        
        mn_list_s = np.asarray( mn_list_s)
        mn_list_s = mn_list_s[  mn_list_s[:,0].argsort()] 
        mn_list_s = np.unique(  mn_list_s, axis = 0 )
        
        mn_list_d = np.asarray( mn_list_d)
        mn_list_d = mn_list_s[  mn_list_d[:,0].argsort()] 
        mn_list_d = np.unique(  mn_list_d, axis = 0 )

        r_s = dict( via = 0 , track = 0) 
        r_d = dict( via = 0 , track = 0) 

        for mn_list, n_track, terminal, r_dict in zip( (mn_list_s,mn_list_d), (n_s, n_d), ( D, S), (r_d, r_s) ):
            via, track = self._route_pattern_via_track( r12, mn_list, n_track, terminal)
            r_dict["via"]   = via
            r_dict["track"] = track 

        # Modify track
        if r_d["track"] == None: 
            mn_sub    = mn_list_d[0]
            mn_sub[1] = n_d
            
            m_ext = routing_map["D_extension0_m"]
            m_extl = 0
            m_extr = 0
            if m_ext[0] != None:
                m_extl = m_ext[0]
                m_extr = m_ext[1]

            r_d["track"]   = r12.route( mn = [ mn_sub + [m_extl,0], mn_sub + [m_extr,0] ], via_tag =[None, None] )

        if r_s["track"] == None: 
            mn_sub    = mn_list_s[0]
            mn_sub[1] = n_s
            
            m_ext = routing_map["S_extension0_m"]
            m_extl = 0
            m_extr = 0
            if m_ext[0] != None:
                m_extl = m_ext[0]
                m_extr = m_ext[1]
            r_s["track"]   = r12.route( mn = [ mn_sub + [m_extl,0], mn_sub + [m_extr,0] ], via_tag =[None, None] )

        nelements["VIA_D"] = r_d["via"]
        nelements["VIA_S"] = r_s["via"]

        nelements["RD0"] = r_d["track"]
        nelements["RS0"] = r_s["track"]

        # Rail routing
        if params["rail"]:

            tinst_l = 0 # left most sub-element name
            tinst_r = 0 # right most sub-element name

            for i, tinst in enumerate(placement_pattern):
                name = placement_pattern[i]
                if name in iparams:
                    if name != "gbndl":
                        tinst_l = name
                        break   
            
            for i, tinst in enumerate(placement_pattern):
                name = placement_pattern[-1 - i]
                if name in iparams:
                    if name != "gbndr":
                        tinst_r = name
                        break   
                    
            inst_l = iparams[tinst_l]
            inst_r = iparams[tinst_r]

            xy_bl = inst_l.bbox[0]
            xy_tr = inst_r.bbox[1]
            xy_br = [xy_tr[0], xy_bl[1] ]

            mn_bl = r12.xy >= xy_bl 
            mn_br = r12.xy <= xy_br 

            r_track = r12.route( mn = [mn_bl, mn_br], via_tag = [None, None] )
            nelements["RRAIL0"] = r_track
            '''
            xy    = icore.bbox
            xy_bl = xy[0]
            xy_tr = xy[1]
            xy_br = [xy_tr[0], xy_bl[1] ]
            mn_bl = r12.xy >= xy_bl 
            mn_br = r12.xy <= xy_br 

            r_track = r12.route( mn = [mn_bl, mn_br], via_tag = [None, None] )
            nelements["RRAIL0"] = r_track
            '''

        # TIE
        if params["tie"] != False:

            if params["tie"] == "TAP0" or params["tie"] == True:
                del nelements["RS0"]
                del nelements["VIA_S"]
                
                via, track = self._route_pattern_via_track(r12, mn_list_s, n_r, "VIA_TIE")
                r_t        = r12.route_via_track( mn = mn_list_s, via_tag = [ False, False], track = [None, n_r])
            
            elif params["tie"] == "TAP1":
                del nelements["RD0"]
                del nelements["VIA_D"]
                
                via, track = self._route_pattern_via_track(r12, mn_list_d, n_r, "VIA_TIE")
                r_t        = r12.route_via_track( mn = mn_list_d, via_tag = [ False, False], track = [None, n_r])

            else:
                raise Exception(" Value error for tie")
            
            nelements["VIA_TIE"] = via

            if r_t[-1] == None:
                del r_t[1]
            for i, r in enumerate(r_t):
                nelements[f"RECT_TIE{i}"] = r

        return nelements

class TileSpaceTemplate(TileTemplate):
    """The class for tile-based spacer template."""

    def _update_params_sub(self, params_in, params):
        params["rail"     ]  = params_in.get("rail", True )  
        return params

    def pins_func(self, params):
        """The method of creating pin."""
        params  = self._update_params(params)
        pins    = dict()
        
        # generate a virtual routing structure for reference
        route_obj = self._internal_route( params )

        if 'RRAIL0' in route_obj:  # rail
            r_obj = route_obj['RRAIL0']
            pins['RAIL'] = laygo2.object.Pin(xy=r_obj.xy, layer=r_obj.layer, netname='RAIL')
        return pins


    def _internal_route(self, params):
        params    = self._update_params(params)
        glib      = self.glib
        routing_gname = self.routing_gname
        routing_map = self.routing_map
        nf_core   = self.nf_core
        iparams   = self._device_place(params)
        placement_pattern = self.placement_pattern

        r12 = glib[routing_gname]
        n_r = routing_map.get( "RAIL", 0 )

        nelements = {}
        nelements.update(iparams)
        
        icore   = iparams["core"]
        nf      = params["nf"]

        # Rail routing
        if params["rail"]:

            tinst_l = 0 # left most sub-element name
            tinst_r = 0 # right most sub-element name

            for i, tinst in enumerate(placement_pattern):
                name = placement_pattern[i]
                if name in iparams:
                    if name != "gbndl":
                        tinst_l = name
                        break   
            
            for i, tinst in enumerate(placement_pattern):
                name = placement_pattern[-1 - i]
                if name in iparams:
                    if name != "gbndr":
                        tinst_r = name
                        break   
                    
            inst_l = iparams[tinst_l]
            inst_r = iparams[tinst_r]

            xy_bl = inst_l.bbox[0]
            xy_tr = inst_r.bbox[1]
            xy_br = [xy_tr[0], xy_bl[1] ]

            mn_bl = r12.xy >= xy_bl 
            mn_br = r12.xy <= xy_br 

            r_track = r12.route( mn = [mn_bl, mn_br], via_tag = [None, None] )
            nelements["RRAIL0"] = r_track
            '''
            xy    = icore.bbox
            xy_bl = xy[0]
            xy_tr = xy[1]
            xy_br = [xy_tr[0], xy_bl[1] ]
            mn_bl = r12.xy >= xy_bl 
            mn_br = r12.xy <= xy_br 

            r_track = r12.route( mn = [mn_bl, mn_br], via_tag = [None, None] )
            nelements["RRAIL0"] = r_track
            '''

        return nelements




import typing as t
from os import PathLike
import numpy as np

if t.TYPE_CHECKING:
    import numpy.typing as npt

T = t.TypeVar("T")
FuncType = t.Callable[..., t.Any]
F = t.TypeVar("F", bound=FuncType)



#!/usr/bin/python
########################################################################################################################
#
# Copyright (C) 2023, Nifty Chips Laboratory at Hanyang University - All Rights Reserved
# 
# Unauthorized copying of this software package, via any medium is strictly prohibited
# Proprietary and confidential
# Written by Jaeduk Han, 07-23-2023
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements the interface with matplotlib.

"""
from typing import (
    TYPE_CHECKING,
    Union,
    List
)

import logging
from math import log10
from decimal import *

import numpy as np
import laygo2.util.transform as tf

import matplotlib
import matplotlib.pyplot as plt

import laygo2.object

#from laygo2._typing import Path

# Type checking
from typing import TYPE_CHECKING, overload, Generic, Dict
from typing import List, Tuple, Iterable, Type, Union, Any, Optional
if TYPE_CHECKING:
    import laygo2

__author__ = ""
__maintainer__ = ""
__status__ = "Prototype"

# support for hover information.
# https://stackoverflow.com/a/47166787/7128154
# https://matplotlib.org/3.3.3/api/collections_api.html#matplotlib.collections.PathCollection
# https://matplotlib.org/3.3.3/api/path_api.html#matplotlib.path.Path
# https://stackoverflow.com/questions/15876011/add-information-to-matplotlib-navigation-toolbar-status-bar
# https://stackoverflow.com/questions/36730261/matplotlib-path-contains-point
# https://stackoverflow.com/a/36335048/7128154
class StatusbarHoverManager:
    """
    Manage hover information for mpl.axes.Axes object based on appearing
    artists.

    Attributes
    ----------
    ax : mpl.axes.Axes
        subplot to show status information
    artists : list of mpl.artist.Artist
        elements on the subplot, which react to mouse over
    labels : list (list of strings) or strings
        each element on the top level corresponds to an artist.
        if the artist has items
        (i.e. second return value of contains() has key 'ind'),
        the element has to be of type list.
        otherwise the element if of type string
    cid : to reconnect motion_notify_event
    """
    def __init__(self, fig, ax):
        assert isinstance(ax, matplotlib.axes.Axes)

        def hover(event):
            if event.inaxes != ax:
                return
            info = 'x={:.0f}, y={:.0f}'.format(event.xdata, event.ydata)
            self.text.set_text(info)
            # replace with self.text.set_text(info) if you want to show the info in the figure information bar.
            #ax.format_coord = lambda x, y: info
        cid = ax.figure.canvas.mpl_connect("motion_notify_event", hover)
        self.ax = ax
        self.cid = cid
        self.artists = []
        self.labels = []
        text = fig.text(0.0, 0.02, "", va="bottom", ha="left", fontsize=6)
        self.fig = fig
        self.text = text

    def add_artist_labels(self, artist, label):
        if isinstance(artist, list):
            assert len(artist) == 1
            artist = artist[0]

        self.artists += [artist]
        self.labels += [label]

        def hover(event):
            if event.inaxes != self.ax:
                return
            info = 'x={:.0f}, y={:.0f}'.format(event.xdata, event.ydata)
            for aa, artist in enumerate(self.artists):
                cont, dct = artist.contains(event)
                if not cont:
                    continue
                inds = dct.get('ind')
                if inds is not None:  # artist contains items
                    for ii in inds:
                        lbl = self.labels[aa][ii]
                        info += '\n   {:}'.format(lbl)
                else:
                    lbl = self.labels[aa]
                    info += '\n   {:}'.format(lbl)
            self.text.set_text(info)
            self.fig.canvas.draw_idle()
            # replace with self.text.set_text(info) if you want to show the info in the figure information bar.
            #self.ax.format_coord = lambda x, y: info

        self.ax.figure.canvas.mpl_disconnect(self.cid)
        self.cid = self.ax.figure.canvas.mpl_connect("motion_notify_event", hover)


def _translate_obj(
        objname: str, 
        obj: "laygo2.object.physical.PhysicalObject", 
        colormap: dict, 
        scale=1, 
        master: "laygo2.object.physical.PhysicalObject" = None, 
        offset=np.array([0, 0])
        ):
    """
    Convert a layout object to corresponding matplotlib patch object.

    Parameters
    ----------
    objname: str
        The name of the object. 
    obj: laygo2.object.PhysicalObject
        The object to be converted.
    colormap: dict
        A dictionary that contains layer-color mapping information.
    scale: float
        (optional) The scaling factor between laygo2's integer coordinates and plot coordinates.
    master: laygo2.object.PhysicalObject
        (optional) The master object of the translated object.
    offset: np.array
        (optional) The offset of the translated object's position.
    
    Returns
    -------
    pypobjs: list
        A list of matplotlib patch objects. Each element is a list of the form [patch, layer, annotation, (on-figure annotation)].
    """
    if master is None:
        mxy = np.array([0, 0])
        mtf = "R0"
    else:  # if the translated object has a master (e.g. VirtualInstance)
        mxy = master.xy
        mtf = master.transform
    if obj.__class__ == laygo2.object.physical.Rect:
        # Compute the rectangle size
        _xy = np.sort(obj.xy, axis=0)  # make sure obj.xy is sorted
        _xy = mxy + np.dot(
            _xy + np.array([[-obj.hextension, -obj.vextension], [obj.hextension, obj.vextension]]),
            tf.Mt(mtf).T,
        )
        size = [_xy[1, 0] - _xy[0, 0], _xy[1, 1] - _xy[0, 1]]
        # Create a rectangle patch
        if obj.layer[0] in colormap:
            rect = matplotlib.patches.Rectangle(
                (_xy[0, 0], _xy[0, 1]),
                size[0],
                size[1],
                facecolor=colormap[obj.layer[0]][1],
                edgecolor=colormap[obj.layer[0]][0],
                alpha=colormap[obj.layer[0]][2],
                lw=2,
            )
            return [[rect, obj.layer[0], "[rect] " + objname + "/" + obj.layer[0] + "/" +str(obj.netname)]]
        return []
    elif obj.__class__ == laygo2.object.physical.Path:
        # TODO: implement path export function.
        pass
    elif obj.__class__ == laygo2.object.physical.Pin:
        if obj.elements is None:
            _objelem = [obj]
        else:
            _objelem = obj.elements
        for idx, _obj in np.ndenumerate(_objelem):
            # Compute the pin rectangle size
            _xy = mxy + np.dot(_obj.xy, tf.Mt(mtf).T)
            size = [_xy[1, 0] - _xy[0, 0], _xy[1, 1] - _xy[0, 1]]
            if obj.layer[0] in colormap:
                # Create a pin rectangle patch
                rect = matplotlib.patches.Rectangle(
                    (_xy[0, 0], _xy[0, 1]),
                    size[0],
                    size[1],
                    facecolor=colormap[obj.layer[0]][1],
                    edgecolor=colormap[obj.layer[0]][0],
                    alpha=colormap[obj.layer[0]][2],
                    lw=2,
                )
                pypobj = [rect, obj.layer[0], "[pin] " + objname + "/" + obj.layer[0] + "/" + str(obj.netname)]
                if master is None:  # add on-figure annotatation if the pin is at the top level
                    pypobj.append(obj.layer[0]+"/"+str(obj.netname))
                return [pypobj]
            return []
    elif obj.__class__ == laygo2.object.physical.Text:
        return [["text", obj.layer[0], obj.text, obj.xy]]
        # TODO: implement text export function.
        pass
    elif obj.__class__ == laygo2.object.physical.Instance:
        # Check if the instance belongs to a virtual instance or array.
        if master is not None:
            _master = master
        elif obj.master is not None:
            _master = obj.master
        else:
            _master = None
        if obj.shape is None:  # single instance
            _xy = mxy + np.dot(obj.xy, tf.Mt(mtf).T)
            _xy0 = _xy
            _xy1 = np.dot(obj.xy1-obj.xy0, tf.Mt(mtf).T) #* np.array([num_rows, num_cols])
            rect = matplotlib.patches.Rectangle(
                (_xy0[0], _xy0[1]),
                _xy1[0],
                _xy1[1],
                facecolor=colormap["__instance__"][1],
                edgecolor=colormap["__instance__"][0],
                alpha=colormap["__instance__"][2],
                lw=2,
            )
            pypobj = [rect, "__instance__", "[inst] " + objname + "/" + obj.cellname]
            if master is None:  # add on-figure annotatation if the instance is at the top level
                if not obj.cellname.startswith("via"):
                    pypobj.append(objname+"/"+obj.cellname)
            pypobjs = [pypobj]
        else: # array instance
            pypobjs = []
            for i, e in np.ndenumerate(obj.elements):
                pypobjs += _translate_obj(e.name+"_"+str(i), e, colormap, scale=scale, master=_master)

        # Instance pins
        for pn, p in obj.pins.items():
            _pypobj = _translate_obj(pn, p, colormap, scale=scale, master=_master, offset=offset)
            _pypobj[0][1] = "__instance_pin__"
            _pypobj[0][0].set(edgecolor=colormap["__instance_pin__"][0])
            _pypobj[0][0].set(facecolor=colormap["__instance_pin__"][1])
            _pypobj[0][2] = "[inst] " + objname + " " + _pypobj[0][2]
            pypobjs += _pypobj
        return pypobjs
    elif obj.__class__ == laygo2.object.physical.VirtualInstance:
        # construct the list that constains patches of elements.
        pypobjs = []
        if obj.shape is None:
            _xy = mxy + np.dot(obj.xy, tf.Mt(mtf).T)
            _xy0 = _xy
            _xy1 = np.dot(obj.xy1-obj.xy0, tf.Mt(mtf).T) 
            rect = matplotlib.patches.Rectangle(
                (_xy0[0], _xy0[1]),
                _xy1[0],
                _xy1[1],
                facecolor=colormap["__instance__"][1],
                edgecolor=colormap["__instance__"][0],
                alpha=colormap["__instance__"][2],
                lw=2,
            )
            pypobjs = [[rect, "__instance__", "[vinst] " + objname + "/" + obj.cellname]]
            pypobjs[0].append(objname+"/"+obj.cellname)
            for elem_name, elem in obj.native_elements.items():
                if not elem.__class__ == laygo2.object.physical.Pin:
                    if obj.name == None:
                        objname = "NoName"
                    else:
                        objname = obj.name
                    _pypobj = _translate_obj(objname, elem, colormap, scale=scale, master=obj)
                    for _p in _pypobj:
                        _p[2] = "[vinst] " + objname + " " + _p[2]
                    pypobjs += _pypobj
        else:  # arrayed VirtualInstance
            for i, j in np.ndindex(tuple(obj.shape.tolist())):  # iterate over obj.shape
                for elem_name, elem in obj.native_elements.items():
                    if not elem.__class__ == laygo2.object.physical.Pin:
                        _pypobj = _translate_obj(
                            obj.name + "_" + elem_name + str(i) + "_" + str(j),
                            elem,
                            colormap,
                            scale=scale,
                            master=obj[i, j],
                        )
                        _pypobj[0][2] = "[vinst] " + objname + " " + _pypobj[0][2]
                        pypobjs += _pypobj

        # Instance pins
        for pn, p in obj.pins.items():
            # master coordinate is already included in the pin object of the virtual instance.
            _pypobj = _translate_obj(pn, p, colormap, scale=scale, master=None, offset=offset)  
            _pypobj[0][1] = "__instance_pin__"
            _pypobj[0][0].set(edgecolor=colormap["__instance_pin__"][0])
            _pypobj[0][0].set(facecolor=colormap["__instance_pin__"][1])
            _pypobj[0][2] = "[vinst] " + objname + " " + _pypobj[0][2]
            pypobjs += _pypobj
        return pypobjs

    else:
        return []
        # return [obj.translate_to_matplotlib()]
    return []

def export(
    db: Union["laygo2.object.database.Library", "laygo2.object.database.Design"],
    cellname: str = None,
    tech: "laygo2.object.technology.BaseTechnology" = None,
    scale = 1,
    colormap: dict = None,
    order = None,
    xlim: list = None,
    ylim: list = None,
    show: bool = False,
    filename: str = None,
    annotate_grid: List["laygo2.object.physical.Grid"] = None,
):
    """
    Export a laygo2.object.database.Library or Design object to a matplotlib plot.

    Parameters
    ----------
    db: laygo2.database.Library or laygo2.database design
        The library database or design to exported.
    cellname: str or List[str]
        (optional) The name(s) of cell(s) to be exported.
    tech: laygo2.technology.BaseTechnology
        (optional) The technology object to be used for the export.
    scale: float
        (optional) The scaling factor between laygo2's integer coordinates and plot coordinates.
    colormap: dict
        A dictionary that contains layer-color mapping information.
    order: list
        A list that contains the order of layers to be displayed (former is plotted first).
    xlim: list
        (optional) A list that specifies the range of plot in x-axis.
    ylim: list
        (optional) A list that specifies the range of plot in y-axis.
    filename: str
        (optional) If specified, export a output file for the plot.
    annotate_grid: list
        (optional) A list of grid objects to be annotated.

    Returns
    -------
    matplotlib.pyplot.figure or list: The generated figure object(s).

    """
    # colormap
    if colormap is None:
        if tech is None:
            colormap = dict()
        else:
            tech_params = tech.tech_params
            colormap = tech_params['export']['mpl']['colormap']

    # a list to align layered objects in order
    if order is None:
        if tech is None:
            order = []
        else:
            tech_params = tech.tech_params
            order = tech_params['export']['mpl']['order']

    # cell name handling.
    if isinstance(db, laygo2.object.database.Design):
        cellname = [db.cellname]
        db = {db.cellname:db}
    if isinstance(db, laygo2.object.database.Library):
        cellname = db.keys() if cellname is None else cellname  # export all cells if cellname is not given.
        cellname = [cellname] if isinstance(cellname, str) else cellname  # convert to a list for iteration.

    fig = []
    _fig = plt.figure()
    for cn in cellname:
        if not show:
            _fig.set_figwidth(max(6, int(6*db[cn].bbox[1, 0]/1200)))
            _fig.set_figheight(max(6, int(6*db[cn].bbox[1, 1]/1200)))
        pypobjs = []
        ax = _fig.add_subplot(111)
        shm = StatusbarHoverManager(_fig, ax)
        for objname, obj in db[cn].items():
            pypobjs += _translate_obj(objname, obj, colormap, scale=scale)
        for _alignobj in order:
            for _pypobj in pypobjs:
                if _pypobj[1] == _alignobj:  # layer is matched.
                    if isinstance(_pypobj[0], str):
                        if _pypobj[0] == 'text':  # Text
                            color = "black"
                            ax.annotate(
                                _pypobj[2], (_pypobj[3][0], _pypobj[3][1]), color=color, weight="bold", fontsize=6, ha="center", va="center"
                            )
                    elif _pypobj[0].__class__ == matplotlib.patches.Rectangle:  # Rect
                        ax.add_patch(_pypobj[0])
                        if len(_pypobj) >= 3:  # annotation.
                            shm.add_artist_labels(_pypobj[0], _pypobj[2])
                            #ax.add_artist(_pypobj[0])
                            if len(_pypobj) == 4:  # text annotation.
                                rx, ry = _pypobj[0].get_xy()
                                cx = rx + _pypobj[0].get_width() / 2.0
                                cy = ry + _pypobj[0].get_height() / 2.0
                                if _pypobj[1] == "__instance__":
                                    color = _pypobj[0].get_edgecolor()
                                elif _pypobj[1] == "__instance_pin__":
                                    color = _pypobj[0].get_edgecolor()
                                else:
                                    color = "black"
                                ax.annotate(
                                    _pypobj[3], (cx, cy), color=color, weight="bold", fontsize=6, ha="center", va="center"
                                )
        if annotate_grid is not None:
            for grid in annotate_grid:
                laygo2.interface.mpl.export_grid(
                    obj=grid,
                    colormap=colormap,
                    order=order,
                    xlim=db[cn].bbox[:, 0],
                    ylim=db[cn].bbox[:, 1],
                    filename=None,
                    fig=_fig,
                    ax=ax,
                    for_annotation=True,
                )
        # scale
        ax.set_aspect('equal', adjustable='box')
        #_fig.tight_layout()
        plt.autoscale()
        if not (xlim == None):
            plt.xlim(xlim)
        if not (ylim == None):
            plt.ylim(ylim)
        fig.append(_fig)
    
    if len(fig) == 1:
        fig = fig[-1]
    #fig = fig[0]

    if filename is not None:
        plt.savefig(filename)

    if show:
        plt.show()
    else:
        plt.clf()
        plt.close()

    return fig

def export_instance(
    obj: "laygo2.object.physical.Instance",
    scale = 1,
    colormap: dict = None,
    order: list = None,
    xlim: list = None,
    ylim: list = None,
    filename: str = None,
):
    """
    Export a laygo2.object.physical.Instance object to a matplotlib plot.

    Parameters
    ----------
    obj: laygo2.object.physical.Instance
        The instance object to exported.
    scale: float
        (optional) The scaling factor between laygo2's integer coordinates and plot coordinates.
    colormap: dict
        A dictionary that contains layer-color mapping information.
    order: list
        A list that contains the order of layers to be displayed (former is plotted first).
    xlim: list
        (optional) A list that specifies the range of plot in x-axis.
    ylim: list
        (optional) A list that specifies the range of plot in y-axis.
    filename: str
        (optional) If specified, export a output file for the plot.

    Returns
    -------
    matplotlib.pyplot.figure or list: The generated figure object(s).

    """

    # colormap
    if colormap is None:
        colormap = dict()

    # a list to align layered objects in order
    if order is None:
        order = []

    '''
    # xlim and ylim
    if xlim is None:
        xlim = [obj.bbox[0][0] - obj.width, obj.bbox[1][0] + obj.width]
    if ylim is None:
        ylim = [obj.bbox[0][1] - obj.height, obj.bbox[1][1] + obj.height]
    '''

    fig = plt.figure()
    pypobjs = []
    ax = fig.add_subplot(111)
    pypobjs += _translate_obj(obj.name, obj, colormap, scale=scale)
    for _alignobj in order:
        for _pypobj in pypobjs:
            if _pypobj[1] == _alignobj:  # layer is matched.
                if isinstance(_pypobj[0], str):
                    if _pypobj[0] == 'text':  # Text
                        # [["text", obj.layer[0], obj.text, obj.xy]]
                        color = "black"
                        ax.annotate(
                            _pypobj[2], (_pypobj[3][0], _pypobj[3][1]), color=color, weight="bold", fontsize=6, ha="center", va="center"
                        )
                elif _pypobj[0].__class__ == matplotlib.patches.Rectangle:  # Rect
                    ax.add_patch(_pypobj[0])
                    if len(_pypobj) == 3:  # annotation.
                        ax.add_artist(_pypobj[0])
                        rx, ry = _pypobj[0].get_xy()
                        cx = rx + _pypobj[0].get_width() / 2.0
                        cy = ry + _pypobj[0].get_height() / 2.0
                        if _pypobj[1] == "__instance_pin__":
                            color = _pypobj[0].get_edgecolor()
                        else:
                            color = "black"
                        ax.annotate(
                            _pypobj[2], (cx, cy), color=color, weight="bold", fontsize=6, ha="center", va="center"
                        )

    # scale
    plt.autoscale()
    if not (xlim is None):
        plt.xlim(xlim)
    if not (ylim is None):
        plt.ylim(ylim)

    if filename is not None:
        plt.savefig(filename)

    plt.show()

    return fig


def export_grid(
    obj: "laygo2.object.physical.Grid",
    colormap: dict = None,
    order: list = None,
    xlim: list = None,
    ylim: list = None,
    filename: str = None,
    fig: matplotlib.pyplot.figure = None,
    ax: matplotlib.pyplot.axis = None,
    for_annotation: bool = False,
):
    """
    Export a laygo2.object.grid.Grid object to a matplotlib plot.

    Parameters
    ----------
    obj: laygo2.object.grid.Grid
        The grid object to exported.
    scale: float
        (optional) The scaling factor between laygo2's integer coordinates and plot coordinates.
    colormap: dict
        A dictionary that contains layer-color mapping information.
    order: list
        A list that contains the order of layers to be displayed (former is plotted first).
    xlim: list
        (optional) A list that specifies the range of plot in x-axis.
    ylim: list
        (optional) A list that specifies the range of plot in y-axis.
    filename: str
        (optional) If specified, export a output file for the plot.
    fig: matplotlib.pyplot.figure
        (optional) The figure object to be plotted on.
    ax: matplotlib.pyplot.axis
        (optional) The axis object to be plotted on.
    for_annotation: bool
        (optional) If True, only draw the dashed grid lines.

    Returns
    -------
    matplotlib.pyplot.figure or list: The generated figure object(s).

    """
    # colormap
    if colormap is None:
        colormap = dict()

    if for_annotation:
        lw = 0.5
        linestyle = "--"
        wscl = 0
    else:
        lw = 2.0
        linestyle = "-"
        wscl = 1

    # a list to align layered objects in order
    if order is None:
        order = []

    if fig is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    pypobjs = []
    # scope
    _xy = (obj.vgrid.range[0], obj.hgrid.range[0])
    _width = obj.vgrid.range[1] - obj.vgrid.range[0]
    _height = obj.hgrid.range[1] - obj.hgrid.range[0]
    rect = matplotlib.patches.Rectangle(_xy, _width, _height, facecolor="none", edgecolor="black", linestyle=linestyle, lw=lw)
    ax.add_patch(rect)
    if not for_annotation:
        rx, ry = rect.get_xy()
        cx = rx + rect.get_width() / 2.0
        cy = ry + rect.get_height() / 2.0
        ax.annotate(
            obj.name, (cx, cy), color="black", weight="bold", fontsize=6, ha="center", va="center"
        )
    if obj.__class__ == laygo2.object.grid.RoutingGrid:  # Routing grid
        if for_annotation:
            rng = range(obj.vgrid<=xlim[0],obj.vgrid>=xlim[1])
            height0 = ylim[1]
        else:
            rng = range(len(obj.vgrid.elements))
            height0 = obj.hgrid.range[1]
        for i in rng:  # vertical routing grid
            ve = obj.vgrid[i][0]
            _xy = (ve - obj.vwidth[i]/2*wscl, obj.hgrid.range[0]-obj.vextension[i])
            _width = obj.vwidth[i]*wscl
            _height = height0 - obj.hgrid.range[0] + 2 * obj.vextension[i]
            facecolor=colormap[obj.vlayer[i][0]][1]
            edgecolor=colormap[obj.vlayer[i][0]][0]
            alpha=colormap[obj.vlayer[i][0]][2]
            rect = matplotlib.patches.Rectangle(_xy, _width, _height, facecolor=facecolor, edgecolor=edgecolor, alpha=alpha, linestyle=linestyle, lw=lw)
            ax.add_patch(rect)
            _xy = (ve, obj.hgrid.range[0]-obj.vextension[i])
            ax.annotate(_xy[0], _xy, color="black", weight="bold", fontsize=6, ha="center", va="center")
        if for_annotation:
            rng = range(obj.hgrid<=ylim[0],obj.hgrid>=ylim[1])
            width0 = xlim[1]
        else:
            rng = range(len(obj.hgrid.elements)) 
            width0 = obj.vgrid.range[1] 
        for i in rng: # horizontal routing grid
            he = obj.hgrid[i][0]
            _xy = (obj.vgrid.range[0] - obj.hextension[i], he - obj.hwidth[i]/2*wscl)
            _width = width0 - obj.vgrid.range[0] + 2*obj.hextension[i]
            _height = obj.hwidth[i]*wscl
            facecolor=colormap[obj.hlayer[i][0]][1]
            edgecolor=colormap[obj.hlayer[i][0]][0]
            alpha=colormap[obj.hlayer[i][0]][2]
            rect = matplotlib.patches.Rectangle(_xy, _width, _height, facecolor=facecolor, edgecolor=edgecolor, alpha=alpha, linestyle=linestyle, lw=lw)
            ax.add_patch(rect)
            _xy = (obj.vgrid.range[0] - obj.hextension[i], he)
            ax.annotate(_xy[1], _xy, color="black", weight="bold", fontsize=6, ha="center", va="center")
        # viamap
        if for_annotation is False:
            for i in range(len(obj.vgrid.elements)):  # vertical routing grid
                for j in range(len(obj.hgrid.elements)):  # horizontal routing grid
                    v = obj.viamap[i, j]
                    x = obj.vgrid.elements[i]
                    y = obj.hgrid.elements[j]
                    circ = matplotlib.patches.Circle((x, y), radius=2, facecolor="black", edgecolor="black") #, **kwargs)
                    ax.add_patch(circ)
                    ax.annotate(v.name, (x+2, y), color="black", fontsize=4, ha="left", va="bottom")

    for _alignobj in order:
        for _pypobj in pypobjs:
            if _pypobj[1] == _alignobj:  # layer is matched.
                if isinstance(_pypobj[0], str):
                    if _pypobj[0] == 'text':  # Text
                        # [["text", obj.layer[0], obj.text, obj.xy]]
                        color = "black"
                        ax.annotate(
                            _pypobj[2], (_pypobj[3][0], _pypobj[3][1]), color=color, weight="bold", fontsize=6, ha="center", va="center"
                        )
                elif _pypobj[0].__class__ == matplotlib.patches.Rectangle:  # Rect
                    ax.add_patch(_pypobj[0])
                    if len(_pypobj) == 3:  # annotation.
                        ax.add_artist(_pypobj[0])
                        rx, ry = _pypobj[0].get_xy()
                        cx = rx + _pypobj[0].get_width() / 2.0
                        cy = ry + _pypobj[0].get_height() / 2.0
                        if _pypobj[1] == "__instance_pin__":
                            color = _pypobj[0].get_edgecolor()
                        else:
                            color = "black"
                        ax.annotate(
                            _pypobj[2], (cx, cy), color=color, weight="bold", fontsize=6, ha="center", va="center"
                        )

    # scale
    plt.autoscale()
    if not for_annotation:
        if not (xlim == None):
            plt.xlim(xlim)
        if not (ylim == None):
            plt.ylim(ylim)

    if filename is not None:
        plt.savefig(filename)

    if not for_annotation:
        plt.show()

    return fig



# -*- coding: utf-8 -*-

"""
Pypi update instruction:
1) update the version number in setup.py.
2) run 'python setup.py sdist bdist_wheel' to build whl.
3) run 'python -m twine upload dist/*'.
4) type username and password.

"""

from setuptools import setup, find_packages

with open("README.md") as f:
    readme = f.read()

with open("LICENSE") as f:
    license = f.read()

setup(
    name="laygo2",
    version="0.6.9",
    author="Jaeduk Han",
    description="LAYout with Gridded Object 2 - A Python Package for Scripted IC Layout Generation Flow",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/niftylab/laygo2",  # project address
    license=license,
    packages=find_packages(exclude=("tests", "docs", "laygo2_tech_quick_start", "laygo2_tech")),
    package_data={
        "": [
            "interface/skill_export.il",
            "interface/magic_export.tcl",
            "interface/gds_default.layermap",
            "examples/laygo2_tech/laygo2_tech.yaml",
            "examples/laygo2_tech/laygo2_tech.layermap",
        ]
    },
    include_package_data=True,
    python_requires=">=3.5",
)



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements interfaces with gds files via gdspy.
"""

import logging
import pprint
from math import log10
from decimal import *
import laygo2.util.transform as tf

import numpy as np

import laygo2.object


def _load_layermap(layermapfile):
    """
    Load layermap information from layermapfile (Foundry techfile can be used)

    Parameters
    ----------
    layermapfile : str
        layermap filename.

        The example file can be found in default.layermap or see below:
        #technology layer information
        #layername  layerpurpose stream# datatype
        text        drawing 100 0
        prBoundary  drawing 101 0
        M1      drawing 50  0
        M1      pin     50  10
        M2      drawing 51  0
        M2      pin     51  10

    Returns
    -------
    dict
        constructed layermap information.

    """
    layermap = dict()
    f = open(layermapfile, "r")
    for line in f:
        tokens = line.split()
        if not len(tokens) == 0:
            if not tokens[0].startswith("#"):
                name = tokens[0]
                # if not layermap.has_key(name):
                if name not in layermap:
                    layermap[name] = dict()
                layermap[name][tokens[1]] = {
                    "layer": int(tokens[2]),
                    "datatype": int(tokens[3]),
                }
    return layermap


def _translate_obj(
    objname,
    obj,
    layermap,
    scale=0.001,
    master=None,
    offset=np.array([0, 0]),
    pin_label_height=0.1,
):
    """
    Convert an object to corresponding skill commands.
    offset : np.array([int, int])
        Offsets to obj.xy
    """
    # import gdspy here to avoid unnecessary C++ compliations for non-gds options.
    import gdspy

    if master is None:
        mxy = np.array([0, 0])
        mtf = "R0"
    else:  # if the translated object has a master (e.g. VirtualInstance)
        mxy = master.xy
        mtf = master.transform

    if obj.__class__ == laygo2.object.Rect:
        ## TODO: add color handling.
        # color = obj.color # coloring function example for skill.

        _xy = np.sort(obj.xy, axis=0)  # make sure obj.xy is sorted
        _xy = mxy + np.dot(
            _xy
            + np.array(
                [[-obj.hextension, -obj.vextension], [obj.hextension, obj.vextension]]
            ),
            tf.Mt(mtf).T,
        )

        l = layermap[obj.layer[0]][obj.layer[1]]
        rect = gdspy.Rectangle((_xy[0, 0], _xy[0, 1]), (_xy[1, 0], _xy[1, 1]), **l)
        return rect
    elif obj.__class__ == laygo2.object.Path:
        # TODO: implement path export function.
        pass
    elif obj.__class__ == laygo2.object.Pin:
        if obj.elements is None:
            _objelem = [obj]
        else:
            _objelem = obj.elements
        item = []
        for idx, _obj in np.ndenumerate(_objelem):
            _xy = mxy + np.dot(_obj.xy, tf.Mt(mtf).T)
            l = layermap[_obj.layer[0]][_obj.layer[1]]
            rect = gdspy.Rectangle((_xy[0, 0], _xy[0, 1]), (_xy[1, 0], _xy[1, 1]), **l)
            _xy_c = 0.5 * (_xy[0, :] + _xy[1, :])
            text = gdspy.Label(
                _obj.netname, _xy_c, "nw", magnification=pin_label_height * 100
            )
            item += [rect, text]
        return item
    elif obj.__class__ == laygo2.object.Text:
        # TODO: implement text export function.
        pass
    elif obj.__class__ == laygo2.object.Instance:
        print(
            "[Warning] laygo2.interface.gdspy: Instance transform is not implemented yet."
        )
        _xy = mxy + np.dot(obj.xy, tf.Mt(mtf).T)
        if master is None:
            transform = obj.transform
        else:  # if the translated object has a master (e.g. VirtualInstance)
            transform = tf.combine(obj.transform, master.transform)
        if obj.shape is None:
            num_rows = 1
            num_cols = 1
            sp_rows = 0
            sp_cols = 0
        else:
            num_rows = obj.shape[1]
            num_cols = obj.shape[0]
            sp_rows = obj.pitch[1]
            sp_cols = obj.pitch[0]
        # if obj.params is None:  # gds cannot handle pcell parameters.
        #    inst_params = "nil"
        # else:
        #    inst_params = _py2skill_inst_params(obj.params['pcell_params'])
        inst = gdspy.CellReference(obj.cellname, _xy)  # , transform)
        return inst
    elif obj.__class__ == laygo2.object.VirtualInstance:
        item = []
        if obj.shape is None:
            for elem_name, elem in obj.native_elements.items():
                if not elem.__class__ == laygo2.object.Pin:
                    if obj.name == None:
                        obj.name = "NoName"
                    else:
                        pass
                    item += [
                        _translate_obj(
                            obj.name + "_" + elem_name,
                            elem,
                            layermap=layermap,
                            master=obj,
                            scale=scale,
                            pin_label_height=pin_label_height,
                        )
                    ]
        else:  # arrayed VirtualInstance
            for i, j in np.ndindex(tuple(obj.shape.tolist())):  # iterate over obj.shape
                for elem_name, elem in obj.native_elements.items():
                    if not elem.__class__ == laygo2.object.Pin:
                        item += [
                            _translate_obj(
                                obj.name + "_" + elem_name + str(i) + "_" + str(j),
                                elem,
                                layermap=layermap,
                                master=obj[i, j],
                                scale=scale,
                                pin_label_height=pin_label_height,
                            )
                        ]
        return item
    return None
    # raise Exception("No corresponding GDS structure for:"+obj.summarize())


def export(
    db,
    filename,
    cellname=None,
    scale=1e-9,
    layermapfile="default.layermap",
    physical_unit=1e-9,
    logical_unit=0.001,
    pin_label_height=0.1,
    svg_filename=None,
    png_filename=None,
):
    """
    Export a laygo2.object.database.Library object to a gds file via gdspy.

    Parameters
    ----------
    db: laygo2.database.Library
        The library database to exported.
    filename: str, optional
        The name of output gds file.
    cellname: str or List[str]
        The name(s) of cell(s) to be exported.
    scale: float
        The scaling factor between laygo2's integer coordinats actual physical coordinates.
    layermapfile : str
        the name of layermap file.
    physical_unit : float, optional
        GDS physical unit.
    logical_unit : float, optional
        GDS logical unit.
    pin_label_height : float, optional
        the height of pin label.
    svg_filename: str, optional
        If specified, it exports a svg file with the specified filename.
    svg_filename: str, optional
        If specified, it exports a png file with the specified filename
        (svg_filename needs to be specified as well).

    Example
    -------
    >>> import laygo2
    >>> from laygo2.object.database import Design
    >>> from laygo2.object.physical import Rect, Pin, Instance, Text
    >>> # Create a design.
    >>> dsn = Design(name="mycell", libname="genlib")
    >>> # Create layout objects.
    >>> r0 = Rect(xy=[[0, 0], [100, 100]], layer=["M1", "drawing"])
    >>> p0 = Pin(xy=[[0, 0], [50, 50]], layer=["M1", "pin"], name="P")
    >>> i0 = Instance(libname="tlib", cellname="t0", name="I0", xy=[0, 0])
    >>> t0 = Text(xy=[[50, 50], [100, 100]], layer=["text", "drawing"], text="T")
    >>> # Add the layout objects to the design object.
    >>> dsn.append(r0)
    >>> dsn.append(p0)
    >>> dsn.append(i0)
    >>> dsn.append(t0)
    >>> #
    >>> # Export to a gds file.
    >>> lib = laygo2.object.database.Library(name="mylib")
    >>> lib.append(dsn)
    >>> laygo2.interface.gds.export(lib, filename="mylayout.gds")
    """
    # Compute scale parameter.
    _scale = round(1 / scale * physical_unit / logical_unit)
    # 1um in phy
    # 1um/1nm = 1000 in laygo2 if scale = 1e-9 (1nm)
    # 1000/1nm*1nm/0.001 = 1000000 in gds if physical_unit = 1e-9 (1nm) and logical_unit = 0.001

    # Load layermap file.
    layermap = _load_layermap(layermapfile)  # load layermap information

    # Construct cellname.
    cellname = (
        db.keys() if cellname is None else cellname
    )  # export all cells if cellname is not given.
    cellname = (
        [cellname] if isinstance(cellname, str) else cellname
    )  # convert to a list for iteration.

    # import gdspy here to avoid unnecessary C++ compliations for non-gds options.
    import gdspy

    # Create library.
    lib = gdspy.GdsLibrary()
    for cn in cellname:
        # Create cell.
        cell = lib.new_cell(cn)
        # Translate objects.
        for objname, obj in db[cn].items():
            tobj = _translate_obj(
                objname,
                obj,
                layermap=layermap,
                scale=_scale,
                pin_label_height=pin_label_height,
            )
            if tobj is not None:
                cell.add(tobj)
    lib.write_gds(filename)
    if svg_filename is not None:
        cell.write_svg(svg_filename)
        if svg_filename is not None:
            # import cairosvg here to avoid unnecessary lib installation for non-gds options.
            import cairosvg

            cairosvg.svg2png(url=svg_filename, write_to=png_filename, scale=1.0)
    # gdspy.LayoutViewer()



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements the interface with matplotlib.

"""
import logging
from math import log10
from decimal import *

import numpy as np
import laygo2.object
import laygo2.util.transform as tf

import matplotlib
import matplotlib.pyplot as plt

__author__ = ""
__maintainer__ = ""
__status__ = "Prototype"


def _translate_obj(objname, obj, colormap, scale=1, master=None, offset=np.array([0, 0])):
    """
    Convert a layout object to corresponding matplotlib patch object.
    """
    if master is None:
        mxy = np.array([0, 0])
        mtf = "R0"
    else:  # if the translated object has a master (e.g. VirtualInstance)
        mxy = master.xy
        mtf = master.transform
    if obj.__class__ == laygo2.object.Rect:
        # Draw a rectangle
        _xy = np.sort(obj.xy, axis=0)  # make sure obj.xy is sorted
        _xy = mxy + np.dot(
            _xy + np.array([[-obj.hextension, -obj.vextension], [obj.hextension, obj.vextension]]),
            tf.Mt(mtf).T,
        )
        size = [_xy[1, 0] - _xy[0, 0], _xy[1, 1] - _xy[0, 1]]
        if obj.layer[0] in colormap:
            rect = matplotlib.patches.Rectangle(
                (_xy[0, 0], _xy[0, 1]),
                size[0],
                size[1],
                facecolor=colormap[obj.layer[0]][1],
                edgecolor=colormap[obj.layer[0]][0],
                alpha=colormap[obj.layer[0]][2],
                lw=2,
            )
            # ax.add_patch(rect)
            return [[rect, obj.layer[0]]]
        return []
    elif obj.__class__ == laygo2.object.Path:
        # TODO: implement path export function.
        pass
    elif obj.__class__ == laygo2.object.Pin:
        if obj.elements is None:
            _objelem = [obj]
        else:
            _objelem = obj.elements
        for idx, _obj in np.ndenumerate(_objelem):
            # Invoke _laygo2_generate_pin(cv, name, layer, bbox) in {header_filename}
            _xy = mxy + np.dot(_obj.xy, tf.Mt(mtf).T)
            size = [_xy[1, 0] - _xy[0, 0], _xy[1, 1] - _xy[0, 1]]
            if obj.layer[0] in colormap:
                rect = matplotlib.patches.Rectangle(
                    (_xy[0, 0], _xy[0, 1]),
                    size[0],
                    size[1],
                    facecolor=colormap[obj.layer[0]][1],
                    edgecolor=colormap[obj.layer[0]][0],
                    alpha=colormap[obj.layer[0]][2],
                    lw=2,
                )
                return [[rect, obj.layer[0], objname + "/" + obj.netname]]
                # ax.add_patch(rect)
            return []
    elif obj.__class__ == laygo2.object.Text:
        return [["text", obj.layer[0], obj.text, obj.xy]]
        # TODO: implement text export function.
        pass
    elif obj.__class__ == laygo2.object.Instance:
        # Invoke _laygo2_generate_instance( cv name libname cellname viewname loc orient num_rows num_cols
        # sp_rows sp_cols params params_order )
        _xy = mxy + np.dot(obj.xy, tf.Mt(mtf).T)

        _xy0 = obj.xy0
        _xy1 = np.dot(obj.size, tf.Mt(mtf).T) #* np.array([num_rows, num_cols])
        rect = matplotlib.patches.Rectangle(
            (_xy0[0], _xy0[1]),
            _xy1[0],
            _xy1[1],
            facecolor=colormap["__instance__"][1],
            edgecolor=colormap["__instance__"][0],
            alpha=colormap["__instance__"][2],
            lw=2,
        )
        #pypobjs = [[rect, "__instance__", obj.cellname + "/" + obj.name]]
        pypobjs = [[rect, "__instance__", obj.cellname + "/" + objname]]

        # Elements for array instances
        if obj.shape is not None:
            for i, e in np.ndenumerate(obj.elements):
                pypobjs += _translate_obj(e.name+"_"+str(i), e, colormap, scale=scale)

        # Instance pins
        for pn, p in obj.pins.items():
            _pypobj = _translate_obj(pn, p, colormap, scale=scale, master=master, offset=offset)
            _pypobj[0][1] = "__instance_pin__"
            _pypobj[0][0].set(edgecolor=colormap["__instance_pin__"][0])
            _pypobj[0][0].set(facecolor=colormap["__instance_pin__"][1])
            pypobjs += _pypobj
        return pypobjs
    elif obj.__class__ == laygo2.object.VirtualInstance:
        pypobjs = []
        if obj.shape is None:
            for elem_name, elem in obj.native_elements.items():
                if not elem.__class__ == laygo2.object.Pin:
                    if obj.name == None:
                        objname = "NoName"
                    else:
                        objname = obj.name
                    _pypobj = _translate_obj(objname, elem, colormap, scale=scale, master=obj)
                    pypobjs += _pypobj
        else:  # arrayed VirtualInstance
            for i, j in np.ndindex(tuple(obj.shape.tolist())):  # iterate over obj.shape
                for elem_name, elem in obj.native_elements.items():
                    if not elem.__class__ == laygo2.object.Pin:
                        _pypobj = _translate_obj(
                            obj.name + "_" + elem_name + str(i) + "_" + str(j),
                            elem,
                            colormap,
                            scale=scale,
                            master=obj[i, j],
                        )
                        pypobjs += _pypobj

        # Instance pins
        for pn, p in obj.pins.items():
            _pypobj = _translate_obj(pn, p, colormap, scale=scale, master=master, offset=offset)
            _pypobj[0][1] = "__instance_pin__"
            _pypobj[0][0].set(edgecolor=colormap["__instance_pin__"][0])
            _pypobj[0][0].set(facecolor=colormap["__instance_pin__"][1])
            pypobjs += _pypobj
        return pypobjs

    else:
        return []
        # return [obj.translate_to_matplotlib()]
    return []


def export(
    db,
    cellname=None,
    scale=1,
    colormap=None,
    order=None,
    xlim=None,
    ylim=None,
    filename=None,
):
    """
    Export a laygo2.object.database.Library or Design object to a matplotlib plot.

    Parameters
    ----------
    db: laygo2.database.Library or laygo2.database design
        The library database or design to exported.
    cellname: str or List[str]
        (optional) The name(s) of cell(s) to be exported.
    scale: float
        (optional) The scaling factor between laygo2's integer coordinates and plot coordinates.
    colormap: dict
        A dictionary that contains layer-color mapping information.
    order: list
        A list that contains the order of layers to be displayed (former is plotted first).
    xlim: list
        (optional) A list that specifies the range of plot in x-axis.
    ylim: list
        (optional) A list that specifies the range of plot in y-axis.
    filename: str
        (optional) If specified, export a output file for the plot.

    Returns
    -------
    matplotlib.pyplot.figure or list: The generated figure object(s).

    """
    # colormap
    if colormap is None:
        colormap = dict()

    # a list to align layered objects in order
    if order is None:
        order = []

    # cell name handling.
    if isinstance(db, laygo2.database.Design):
        cellname = [db.cellname]
        db = {db.cellname:db}
    if isinstance(db, laygo2.database.Library):
        cellname = db.keys() if cellname is None else cellname  # export all cells if cellname is not given.
        cellname = [cellname] if isinstance(cellname, str) else cellname  # convert to a list for iteration.

    fig = []
    for cn in cellname:
        _fig = plt.figure()
        pypobjs = []
        ax = _fig.add_subplot(111)
        for objname, obj in db[cn].items():
            pypobjs += _translate_obj(objname, obj, colormap, scale=scale)
        for _alignobj in order:
            for _pypobj in pypobjs:
                if _pypobj[1] == _alignobj:  # layer is matched.
                    if isinstance(_pypobj[0], str):
                        if _pypobj[0] == 'text':  # Text
                            # [["text", obj.layer[0], obj.text, obj.xy]]
                            color = "black"
                            ax.annotate(
                                _pypobj[2], (_pypobj[3][0], _pypobj[3][1]), color=color, weight="bold", fontsize=6, ha="center", va="center"
                            )
                    elif _pypobj[0].__class__ == matplotlib.patches.Rectangle:  # Rect
                        ax.add_patch(_pypobj[0])
                        if len(_pypobj) == 3:  # annotation.
                            ax.add_artist(_pypobj[0])
                            rx, ry = _pypobj[0].get_xy()
                            cx = rx + _pypobj[0].get_width() / 2.0
                            cy = ry + _pypobj[0].get_height() / 2.0
                            if _pypobj[1] == "__instance_pin__":
                                color = _pypobj[0].get_edgecolor()
                            else:
                                color = "black"
                            ax.annotate(
                                _pypobj[2], (cx, cy), color=color, weight="bold", fontsize=6, ha="center", va="center"
                            )
        fig.append(_fig)
    if len(fig) == 1:
        fig = fig[0]

    # scale
    plt.autoscale()
    if not (xlim == None):
        plt.xlim(xlim)
    if not (ylim == None):
        plt.ylim(ylim)


    if filename is not None:
        plt.savefig(filename)

    plt.show()

    return fig


def export_instance(
    obj,
    scale=1,
    colormap=None,
    order=None,
    xlim=None,
    ylim=None,
    filename=None,
):
    """
    Export a laygo2.object.physical.Instance object to a matplotlib plot.

    Parameters
    ----------
    obj: laygo2.object.physical.Instance
        The instance object to exported.
    scale: float
        (optional) The scaling factor between laygo2's integer coordinates and plot coordinates.
    colormap: dict
        A dictionary that contains layer-color mapping information.
    order: list
        A list that contains the order of layers to be displayed (former is plotted first).
    xlim: list
        (optional) A list that specifies the range of plot in x-axis.
    ylim: list
        (optional) A list that specifies the range of plot in y-axis.
    filename: str
        (optional) If specified, export a output file for the plot.

    Returns
    -------
    matplotlib.pyplot.figure or list: The generated figure object(s).

    """

    # colormap
    if colormap is None:
        colormap = dict()

    # a list to align layered objects in order
    if order is None:
        order = []

    '''
    # xlim and ylim
    if xlim is None:
        xlim = [obj.bbox[0][0] - obj.width, obj.bbox[1][0] + obj.width]
    if ylim is None:
        ylim = [obj.bbox[0][1] - obj.height, obj.bbox[1][1] + obj.height]
    '''

    fig = plt.figure()
    pypobjs = []
    ax = fig.add_subplot(111)
    pypobjs += _translate_obj(obj.name, obj, colormap, scale=scale)
    for _alignobj in order:
        for _pypobj in pypobjs:
            if _pypobj[1] == _alignobj:  # layer is matched.
                if isinstance(_pypobj[0], str):
                    if _pypobj[0] == 'text':  # Text
                        # [["text", obj.layer[0], obj.text, obj.xy]]
                        color = "black"
                        ax.annotate(
                            _pypobj[2], (_pypobj[3][0], _pypobj[3][1]), color=color, weight="bold", fontsize=6, ha="center", va="center"
                        )
                elif _pypobj[0].__class__ == matplotlib.patches.Rectangle:  # Rect
                    ax.add_patch(_pypobj[0])
                    if len(_pypobj) == 3:  # annotation.
                        ax.add_artist(_pypobj[0])
                        rx, ry = _pypobj[0].get_xy()
                        cx = rx + _pypobj[0].get_width() / 2.0
                        cy = ry + _pypobj[0].get_height() / 2.0
                        if _pypobj[1] == "__instance_pin__":
                            color = _pypobj[0].get_edgecolor()
                        else:
                            color = "black"
                        ax.annotate(
                            _pypobj[2], (cx, cy), color=color, weight="bold", fontsize=6, ha="center", va="center"
                        )
    # scale
    plt.autoscale()
    if not (xlim == None):
        plt.xlim(xlim)
    if not (ylim == None):
        plt.ylim(ylim)

    if filename is not None:
        plt.savefig(filename)

    plt.show()

    return fig


def export_grid(
    obj,
    colormap=None,
    order=None,
    xlim=None,
    ylim=None,
    filename=None,
):
    """
    Export a laygo2.object.grid.Grid object to a matplotlib plot.

    Parameters
    ----------
    obj: laygo2.object.grid.Grid
        The grid object to exported.
    scale: float
        (optional) The scaling factor between laygo2's integer coordinates and plot coordinates.
    colormap: dict
        A dictionary that contains layer-color mapping information.
    order: list
        A list that contains the order of layers to be displayed (former is plotted first).
    xlim: list
        (optional) A list that specifies the range of plot in x-axis.
    ylim: list
        (optional) A list that specifies the range of plot in y-axis.
    filename: str
        (optional) If specified, export a output file for the plot.

    Returns
    -------
    matplotlib.pyplot.figure or list: The generated figure object(s).

    """
    # colormap
    if colormap is None:
        colormap = dict()

    # a list to align layered objects in order
    if order is None:
        order = []

    fig = plt.figure()
    pypobjs = []
    ax = fig.add_subplot(111)
    # scope
    _xy = (obj.vgrid.range[0], obj.hgrid.range[0])
    _width = obj.vgrid.range[1] - obj.vgrid.range[0]
    _height = obj.hgrid.range[1] - obj.hgrid.range[0]
    rect = matplotlib.patches.Rectangle(_xy, _width, _height, facecolor="none", edgecolor="black", lw=2)
    ax.add_patch(rect)
    rx, ry = rect.get_xy()
    cx = rx + rect.get_width() / 2.0
    cy = ry + rect.get_height() / 2.0
    ax.annotate(
        obj.name, (cx, cy), color="black", weight="bold", fontsize=6, ha="center", va="center"
    )
    if obj.__class__ == laygo2.object.RoutingGrid:  # Routing grid
        for i in range(len(obj.vgrid.elements)):  # vertical routing grid
            ve = obj.vgrid.elements[i]
            _xy = (ve - obj.vwidth[i]/2, obj.hgrid.range[0]-obj.vextension[i])
            _width = obj.vwidth[i]
            _height = obj.hgrid.range[1] - obj.hgrid.range[0] + 2 * obj.vextension[i]
            facecolor=colormap[obj.vlayer[i][0]][1]
            edgecolor=colormap[obj.vlayer[i][0]][0]
            alpha=colormap[obj.vlayer[i][0]][2]
            rect = matplotlib.patches.Rectangle(_xy, _width, _height, facecolor=facecolor, edgecolor=edgecolor, alpha=alpha, lw=2)
            ax.add_patch(rect)
            _xy = (ve, obj.hgrid.range[0]-obj.vextension[i])
            ax.annotate(_xy[0], _xy, color="black", weight="bold", fontsize=6, ha="center", va="center")
        for i in range(len(obj.hgrid.elements)):  # horizontal routing grid
            he = obj.hgrid.elements[i]
            _xy = (obj.vgrid.range[0] - obj.hextension[i], he - obj.hwidth[i]/2)
            _width = obj.vgrid.range[1] - obj.vgrid.range[0] + 2*obj.hextension[i]
            _height = obj.hwidth[i]
            facecolor=colormap[obj.hlayer[i][0]][1]
            edgecolor=colormap[obj.hlayer[i][0]][0]
            alpha=colormap[obj.hlayer[i][0]][2]
            rect = matplotlib.patches.Rectangle(_xy, _width, _height, facecolor=facecolor, edgecolor=edgecolor, alpha=alpha, lw=2)
            ax.add_patch(rect)
            _xy = (obj.vgrid.range[0] - obj.hextension[i], he)
            ax.annotate(_xy[1], _xy, color="black", weight="bold", fontsize=6, ha="center", va="center")
        # viamap
        for i in range(len(obj.vgrid.elements)):  # vertical routing grid
            for j in range(len(obj.hgrid.elements)):  # horizontal routing grid
                v = obj.viamap[i, j]
                x = obj.vgrid.elements[i]
                y = obj.hgrid.elements[j]
                circ = matplotlib.patches.Circle((x, y), radius=2, facecolor="black", edgecolor="black") #, **kwargs)
                ax.add_patch(circ)
                ax.annotate(v.name, (x+2, y), color="black", fontsize=4, ha="left", va="bottom")

    for _alignobj in order:
        for _pypobj in pypobjs:
            if _pypobj[1] == _alignobj:  # layer is matched.
                if isinstance(_pypobj[0], str):
                    if _pypobj[0] == 'text':  # Text
                        # [["text", obj.layer[0], obj.text, obj.xy]]
                        color = "black"
                        ax.annotate(
                            _pypobj[2], (_pypobj[3][0], _pypobj[3][1]), color=color, weight="bold", fontsize=6, ha="center", va="center"
                        )
                elif _pypobj[0].__class__ == matplotlib.patches.Rectangle:  # Rect
                    ax.add_patch(_pypobj[0])
                    if len(_pypobj) == 3:  # annotation.
                        ax.add_artist(_pypobj[0])
                        rx, ry = _pypobj[0].get_xy()
                        cx = rx + _pypobj[0].get_width() / 2.0
                        cy = ry + _pypobj[0].get_height() / 2.0
                        if _pypobj[1] == "__instance_pin__":
                            color = _pypobj[0].get_edgecolor()
                        else:
                            color = "black"
                        ax.annotate(
                            _pypobj[2], (cx, cy), color=color, weight="bold", fontsize=6, ha="center", va="center"
                        )

    # scale
    plt.autoscale()
    if not (xlim == None):
        plt.xlim(xlim)
    if not (ylim == None):
        plt.ylim(ylim)

    if filename is not None:
        plt.savefig(filename)

    plt.show()

    return fig



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
Utility functions for coordinate tranformations.
"""

__author__ = "Jaeduk Han"
__maintainer__ = "Jaeduk Han"
__status__ = "Prototype"

import numpy as np

def combine(transform1, transform2):
    """
    Returns the resulting transform parameter of 
    two consecutive transforms
    """
    if transform1 == 'R0':
        if transform2 == 'R0': return 'R0'
        elif transform2 == 'MX': return 'MX'
        elif transform2 == 'MY': return 'MY'
        elif transform2 == 'MXY': return 'MXY'
    elif transform1 == 'MX':
        if transform2 == 'R0': return 'MX'
        elif transform2 == 'MX': return 'R0'
        elif transform2 == 'MY': return 'MXY'
        elif transform2 == 'MXY': return 'MY'
    elif transform1 == 'MY':
        if transform2 == 'R0': return 'MY'
        elif transform2 == 'MX': return 'R180'
        elif transform2 == 'MY': return 'R0'
        elif transform2 == 'MXY': return 'MX'
    raise ValueError("Transformation mapping is not matched.")


def Mt(transform):
    """
    Returns the transform matrix.

    Parameters
    ----------
    transform : str
        The transform parameter. Possible values are 'R0', 'MX', 'MY', 'MXY', and 'R180'.

    Returns
    -------
    numpy.ndarray(dtype=int)
        The transform matrix corresponding to the transform parameter.

    """
    transform_map = {
        'R0': np.array([[1, 0], [0, 1]]),
        'R90': np.array([[0, -1], [1, 0]]),
        'R180': np.array([[-1, 0], [0, -1]]),
        'R270': np.array([[0, 1], [-1, 0]]),
        'MX': np.array([[1, 0], [0, -1]]),
        'MY': np.array([[-1, 0], [0, 1]]),
        'MXY': np.array([[0, 1], [1, 0]]),  # mirror to the y=x line.
    }
    return transform_map[transform]


def Mtinv(transform):
    """
    Returns the inverse of the transform matrix.

    Parameters
    ----------
    transform : str
        The transform parameter. possible values are 'R0', 'MX', 'MY', 'MXY', and 'R180'.

    Returns
    -------
    numpy.ndarray(dtype=int)
        The inverse of the transform matrix.
    """
    transform_map = {
        'R0': np.array([[1, 0], [0, 1]]),    'MX': np.array([[1, 0], [0, -1]]),
        'MY': np.array([[-1, 0], [0, 1]]),
        'MXY': np.array([[0, 1], [1, 0]]),  # mirror to the y=x line.
        'R180': np.array([[-1, 0], [0, -1]]),
    }
    return transform_map[transform]


def Md(direction):
    """
    Returns the direction(projection) matrix. The direction matrix is used when placing an object based on relative
    information to other instance(s). For example, if an instance's center is located at xyc0=[xc0, yc0],
    the xy-coordinate of the center of the new instance xyc1 can be computed from the following equation:

    (1) xyc1 = xyc0 + 0.5 * Md * (xys0 + xys1)

    where xys0, xys1 are the size of the reference and the new instance, respectively, and Md is the direction matrix
    corresponding to the direction of the placement.

    Parameters
    ----------
    direction : str
        The direction parameter. Possible values are 'left', 'right', 'top', 'bottom', 'omni', 'x', 'y'.

    Returns
    -------
    np.array([[int, int], [int, int]])
        The direction matrix.

    Notes
    -----
    The following equation will be used instead of (1) in the future versions, to avoid the 0.5 scaling that increases
    the precision requirement.

    (2) xy1 = xy0 + 0.5 * [(Md + Mt0) * xys0 + (Md - Mt1) * xys1]
    """
    direction_map = {
        'left': np.array([[-1, 0], [0, 0]]),
        'right': np.array([[1, 0], [0, 0]]),
        'top': np.array([[0, 0], [0, 1]]),
        'bottom': np.array([[0, 0], [0, -1]]),
        'omni': np.array([[1, 0], [0, 1]]),  # omnidirectional
        'x': np.array([[1, 0], [0, 0]]),
        'y': np.array([[0, 0], [0, 1]]),
    }
    return direction_map[direction]





#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################
"""
This module implements interface with BAG in skill language
"""
import laygo2.interface
import yaml
import numpy as np

def export(
    db, filename, cellname=None, scale=1e-3, reset_library=False, tech_library=None
):
    """
    Export a laygo2.object.database.Library object to BAG2.

    Parameters
    ----------
    db: laygo2.object.database.Library
        The library database to be exported.
    filename: str, optional
        The path of the intermediate skill script file.
    cellname: str or List[str]
        The name(s) of cell(s) to be exported.
    scale: float
        The scaling factor between laygo2's integer coordinats actual physical coordinates.
    reset_library: bool, optional
        If True, the library to export the cells is reset.
    tech_library: str, optional
        The name of technology library to be attached to the resetted library.

    Returns
    -------
    str: The generated skill script.

    Example
    -------
    >>> import laygo2
    >>> from laygo2.object.database import Design
    >>> from laygo2.object.physical import Rect, Pin, Instance, Text
    >>> # Create a design.
    >>> dsn = Design(name="mycell", libname="genlib")
    >>> # Create layout objects.
    >>> r0 = Rect(xy=[[0, 0], [100, 100]], layer=["M1", "drawing"])
    >>> p0 = Pin(xy=[[0, 0], [50, 50]], layer=["M1", "pin"], name="P")
    >>> i0 = Instance(libname="tlib", cellname="t0", name="I0", xy=[0, 0])
    >>> t0 = Text(xy=[[50, 50], [100, 100]], layer=["text", "drawing"], text="T")
    >>> # Add the layout objects to the design object.
    >>> dsn.append(r0)
    >>> dsn.append(p0)
    >>> dsn.append(i0)
    >>> dsn.append(t0)
    >>> #
    >>> # Export to BAG.
    >>> lib = laygo2.object.database.Library(name="mylib")
    >>> lib.append(dsn)
    >>> scr = laygo2.interface.bag.export(lib, filename="myscript.il")
    >>> print(scr)
    ; (definitions of laygo2 skill functions)
    ; exporting mylib__mycell
    cv = _laygo2_open_layout("mylib" "mycell" "layout")
    _laygo2_generate_rect(cv, list( "M1" "drawing" ), list( list( 0.0000  0.0000  ) list( 0.1000  0.1000  ) ), "None")
    _laygo2_generate_pin(cv, "P", list( "M1" "pin" ), list( list( 0.0000  0.0000  ) list( 0.0500  0.0500  ) ) )
    _laygo2_generate_instance(cv, "I0", "tlib", "t0", "layout", list( 0.0000  0.0000  ), "R0", 1, 1, 0, 0, nil, nil)
    _laygo2_save_and_close_layout(cv)

    """
    skill_str = laygo2.interface.skill.export(
        db, filename, cellname, scale, reset_library, tech_library
    )
    import bag

    prj = bag.BagProject()
    prj.impl_db._eval_skill('load("' + filename + '");1\n')
    print("Your design was generated in Virtuoso.")

    return skill_str

def load(libname, cellname=None, il_filename="import_skill_scratch.il", yaml_filename="import_skill_scratch.yaml", scale=1e-3, mpt=False):
    """
    Import virtuoso layout to a laygo2.object.database.Library object via BAG2 interface.
    """
    import bag
    prj = bag.BagProject()

    if cellname==None: #import all cells
        skill_str = laygo2.interface.skill.load_cell_list(libname, filename=il_filename, yaml_filename=yaml_filename)
        prj.impl_db._eval_skill('load("' + il_filename + '");1\n')
        with open(yaml_filename, 'r') as stream:
            ydict = yaml.load(stream, Loader=yaml.FullLoader)
            celllist=ydict[libname]
    else:
        if isinstance(cellname, list): celllist=cellname
        else: celllist=[cellname]
    
    db = laygo2.object.database.Library(name=libname)
    for cn in celllist:
        skill_str = laygo2.interface.skill.load(libname, cellname=cn, filename=il_filename, 
                                                  yaml_filename=yaml_filename, mpt=mpt)
        prj.impl_db._eval_skill('load("' + il_filename + '");1\n')
        with open(yaml_filename, 'r') as stream:
            ydict = yaml.load(stream, Loader=yaml.FullLoader)
        dsn = laygo2.object.database.Design(name=cn)
        db.append(dsn)
        for _r_key, _r in ydict['rects'].items():
            if 'color' in _r:  # to support MPT
                _color = _r['color']
            else:
                _color = None
            r = laygo2.object.Rect(xy=np.array(_r['bBox'])*(1/scale), layer=_r['layer'].split(), color = _color) # former definition "xy0 / scale" sometimes returns error.
            dsn.append(r)
        for _t_key, _t in ydict['labels'].items():
            t = laygo2.object.Pin(name=_t['label'], xy=np.array(_t['xy'])*(1/scale), layer=_t['layer'].split())
            dsn.append(t)
        for _i_key, _i in ydict['instances'].items():
            if not 'rows' in _i: _i['rows']=1
            if not 'cols' in _i: _i['cols']=1
            if not 'sp_rows' in _i: _i['sp_rows']=0
            if not 'sp_cols' in _i: _i['sp_cols']=0
            if not 'transform' in _i: _i['transform']='R0'
            inst = laygo2.object.Instance(xy = np.array(_i['xy'])*(1/scale), libname = _i['lib_name'], cellname = _i['cell_name'],
                                          shape = np.array([_i['cols'], _i['rows']]), pitch = np.array([_i['sp_cols'], _i['sp_rows']]),
                                          transform = _i['transform']) 
            dsn.append(inst)
    return db


